#define BENCHMARK_ITERATIONS 32

typedef struct BENCHMARK_RESULT benchmark_result;
struct BENCHMARK_RESULT {
    u64 average_ticks;
    u64 min_ticks;
    u64 max_ticks;
    u64 median_ticks;
    u64 ticks_25;
    u64 ticks_75;
    u64 ticks_99;
};

benchmark_result CalculateBenchmarkResult(memory_arena *arena, program_times *times) {
    benchmark_result result = { 0 };

    result.min_ticks = U64_MAX;

    u64 ticks;
    u64 total_ticks = 0;

    u64 *sorted_ticks = GetMemoryArenaArray(arena, u64, BENCHMARK_ITERATIONS);
    for (u32 i = 0; i < BENCHMARK_ITERATIONS; i++) {
        profiler_zone *zone = &times[i].overall;
        ticks = zone->ticks_end - zone->ticks_start;

        total_ticks += ticks;

        // find the index to put the data in...
        u32 index = 0;
        for (;index < i; index++) {
            if (sorted_ticks[index] > ticks) {
                break;
            }
        }

        //...move everything from that index one slot away...
        if (index < i) {
            memmove(&sorted_ticks[index + 1], &sorted_ticks[index], sizeof(*sorted_ticks) * i);
        }

        // ...and insert
        sorted_ticks[index] = ticks;
    }

    // Calculate percentiles
    u32 imin = 0;
    u32 i25 = (u32)(0.25 * (f64)BENCHMARK_ITERATIONS);
    u32 i50 = (u32)(0.5 * (f64)BENCHMARK_ITERATIONS);
    u32 i75 = (u32)(0.75 * (f64)BENCHMARK_ITERATIONS);
    u32 i99 = (u32)(0.99 * (f64)BENCHMARK_ITERATIONS);
    u32 imax = BENCHMARK_ITERATIONS - 1;
    result.min_ticks = sorted_ticks[imin];
    result.ticks_25 = sorted_ticks[i25];
    result.median_ticks = sorted_ticks[i50];
    result.ticks_75 = sorted_ticks[i75];
    result.ticks_99 = sorted_ticks[i99];
    result.max_ticks = sorted_ticks[imax];
    result.average_ticks = total_ticks / BENCHMARK_ITERATIONS;

    return (result);
}

u32 RunBenchmark(memory_arena *arena) {
    InitProgramTimings();

    string outputs[ATTEMPT_COUNT] = {0};

    for (u32 i = 0; i < ATTEMPT_COUNT; i++) {
        string name = GetAttemptName((attempt_enum)i);

        u64 total_pos = GetMemoryArenaPosition(arena);

        program_times times = {0};
        // Warmup
        for (u32 x = 0; x < 4; x++) {
            u64 pos = GetMemoryArenaPosition(arena);
            string wamup_message = StringFormat(arena, "%.*s - Warmup #%lu\n", StringToFormatArgs(name), x);
            WriteMessage(wamup_message);

            RunAttempt((attempt_enum)i, ATTEMPT_TYPE_BENCHMARK, arena, &times);
            RewindMemoryArenaPosition(arena, pos);
        }

        program_times timings[BENCHMARK_ITERATIONS] = {0};
        for (u32 x = 0; x < BENCHMARK_ITERATIONS; x++) {
            u64 pos = GetMemoryArenaPosition(arena);
            string bench_message = StringFormat(arena, "%.*s - Benchmark Run #%lu\n", StringToFormatArgs(name), x);
            WriteMessage(bench_message);

            RunAttempt((attempt_enum)i, ATTEMPT_TYPE_BENCHMARK, arena, &timings[x]);
            RewindMemoryArenaPosition(arena, pos);
        }

        benchmark_result result = CalculateBenchmarkResult(arena, timings);
        RewindMemoryArenaPosition(arena, total_pos);

        f64 average_s = (f64)result.average_ticks / (f64)g_cpu_frequency_in_ticks;
        f64 average_ms = average_s * 1000.0;
        f64 average_us = average_ms * 1000.0;

        outputs[i] = StringFormat(arena,
            "%.*s - Average %$.5llu ticks (%fs - %fms - %fus)(min: %$.5llu, 25%%: %$.5llu, median: %$.5llu, 75%%: %$.5llu, 99%%: %$.5llu, max: %$.5llu)\n",
            StringToFormatArgs(name),
            result.average_ticks,
            average_s,
            average_ms,
            average_us,
            result.min_ticks,
            result.ticks_25,
            result.median_ticks,
            result.ticks_75,
            result.ticks_99,
            result.max_ticks);
        WriteMessage(outputs[i]);
    }

    WriteMessage(StringLiteral("Benchmarks done!\n\n\n"));

    for (u32 i = 0; i < ATTEMPT_COUNT; i++) {
        WriteMessage(outputs[i]);
    }

    return EXIT_CODE_SUCCESS;
}