#ifndef __DANI_LIB_MEMORY_ARENA_H
#define __DANI_LIB_MEMORY_ARENA_H

#include <Windows.h>
#include "dani_base.h"

typedef struct MEMORY_ARENA memory_arena;
struct MEMORY_ARENA {
    memory_arena *previous;
    memory_arena *current;

    u64 reserved;
    u64 committed;
    u64 alignment;

    u64 position;
    u64 base_position;

    u64 max_used;
    b32 is_fixed;
};

#define MEMORY_ARENA_TYPE_SIZE sizeof(memory_arena)

// Arena functions
memory_arena * MemoryArena(u64 reserved, u64 committed, u64 alignment);
void DeleteMemoryArena(memory_arena *arena);

// Absolute position functions
u64 GetMemoryArenaPosition(memory_arena *arena);
void RewindMemoryArenaPosition(memory_arena *arena, u64 new_position);

// Chunk functions
void * GetMemoryArenaChunk(memory_arena *arena, u64 chunk_size);
void ReturnMemoryArenaChunk(memory_arena *arena, u64 chunk_size);

// Helper macros
#define GetMemoryArenaType(arena_ptr, type) ((type *)GetMemoryArenaChunk((arena_ptr), sizeof(type)))
#define ReturnMemoryArenaType(arena_ptr, type) (ReturnMemoryArenaChunk((arena_ptr), sizeof(type)))

#define GetMemoryArenaArray(arena_ptr, type, count) (GetMemoryArenaChunk((arena_ptr), sizeof(type) * (count)))
#define ReturnMemoryArenaArray(arena_ptr, type, count) (ReturnMemoryArenaChunk((arena_ptr), sizeof(type) * (count)))

#endif //__DANI_LIB_MEMORY_ARENA_H

#ifdef DANI_LIB_MEMORY_ARENA_IMPLEMENTATION

// Panic macros
#ifndef DANI_LIB_MEMORY_ARENA_PANIC
#define DANI_LIB_MEMORY_ARENA_PANIC
static void MemoryArenaPanic(void) {
    u32 exit_code = GetLastError();
    ExitProcess(exit_code);
}
#endif // DANI_LIB_MEMORY_ARENA_PANIC

memory_arena * MemoryArena(u64 reserved, u64 committed, u64 alignment) {
    Assert(committed > MEMORY_ARENA_TYPE_SIZE && committed <= reserved);
    Assert(alignment != 0 && IsPower2(alignment));

    SYSTEM_INFO info = {0};
    GetSystemInfo(&info);

    u64 page_size = info.dwPageSize;

    reserved = AlignPower2(reserved, page_size);
    committed = AlignPower2(committed, page_size);

    void *base = VirtualAlloc(0, reserved, MEM_RESERVE, PAGE_READWRITE);
    if (VirtualAlloc(base, committed, MEM_COMMIT, PAGE_READWRITE) == 0) {
        VirtualFree(base, 0, MEM_RELEASE);
        base = 0;
    }

    memory_arena *result = (memory_arena *)base;
    if (result != 0) {
        result->previous = 0;
        result->current = result;

        result->reserved = reserved;
        result->committed = committed;
        result->alignment = alignment;

        result->position = MEMORY_ARENA_TYPE_SIZE;
        result->max_used = result->position;
        result->base_position = 0;
        result->is_fixed = B32_FALSE;
    }

    if (result == 0) {
        MemoryArenaPanic();
    }
    return (result);
}

void DeleteMemoryArena(memory_arena *arena) {
    memory_arena *current = (arena != 0) ? arena->current : 0;
    memory_arena *previous;

    while (current != 0) {
        previous = current->previous;
        VirtualFree(current, 0, MEM_RELEASE);
        current = previous;
    }
}

u64 GetMemoryArenaPosition(memory_arena *arena) {
    u64 result = 0;
    if (arena != 0) {
        memory_arena *current = arena->current;
        result = current->base_position + current->position;
    }
    return (result);
}

u64 GetMemoryArenaMaxUsed(memory_arena *arena) {
    u64 result = (arena != 0) ? arena->current->max_used : 0;
    return (result);
}

void RewindMemoryArenaPosition(memory_arena *arena, u64 new_position) {
    u64 current_position = GetMemoryArenaPosition(arena);
    if (new_position >= current_position) {
        return;
    }

    new_position = ClampFloor(MEMORY_ARENA_TYPE_SIZE, new_position);

    memory_arena *current = arena->current;
    memory_arena *previous;

    while (current->base_position >= new_position) {
        previous = current->previous;
        if (previous->max_used < current->max_used) {
            previous->max_used = current->max_used;
        }

        // MEM_RELEASE because any additional memory arena was not part of the initial reserved space.
        VirtualFree(current, 0, MEM_RELEASE);
        current = previous;
    }

    arena->current = current;

    // Relative new position in the current arena
    new_position = new_position - current->base_position;
    AssertAlways(new_position <= current->position);

    current->position = new_position;
}

void * GetMemoryArenaChunk(memory_arena *arena, u64 chunk_size) {
    memory_arena *current = arena->current;

    u64 start = AlignPower2(current->position, current->alignment);
    u64 end = start + chunk_size;

    if (current->reserved < end && IsFalse(current->is_fixed)) {
        // Data does not fit within the reserved space of the current arena
        memory_arena *new_arena;
        u64 new_arena_size = chunk_size + MEMORY_ARENA_TYPE_SIZE + current->alignment;

        // TODO: Introduce some form of threshold maybe?
        if (current->reserved < new_arena_size) { 
            // New arena needs more space then the previous reserved space
            new_arena = MemoryArena(new_arena_size, new_arena_size, current->alignment);
        } else {
            // New arena can be the same reserved size as the previous one
            new_arena = MemoryArena(current->reserved, new_arena_size, current->alignment);
        }

        if (new_arena != 0) {
            new_arena->base_position = current->base_position + current->reserved;
            new_arena->previous = current;
            arena->current = new_arena;

            current = new_arena;
            start = AlignPower2(current->position, current->alignment);
            end = start + chunk_size;
        }
    }

    if (current->committed < end) {
        // Data does not fit in committed, but might fit in reserved arena space
        SYSTEM_INFO info = {0};
        GetSystemInfo(&info);

        u64 page_size = info.dwPageSize;

        u64 aligned_end = AlignPower2(end, page_size);
        u64 clamped_end = ClampCeiling(aligned_end, current->reserved);
        u64 new_size = clamped_end - current->committed;

        // If new arena allocation failed above this will also fail as we try to commit unreserved pages
        if (VirtualAlloc((u8 *)current + current->committed, new_size, MEM_COMMIT, PAGE_READWRITE) != 0) {
            current->committed = clamped_end;
        }
    }

    void *result = 0;
    if (current->committed >= end) {
        result = (u8 *)current + start;
        current->position = end;

        if (current->base_position + current->position > current->max_used) {
            current->max_used = current->base_position + current->position;
        }
    }

    if (result == 0) {
        MemoryArenaPanic();
    }
    return result;
}

void ReturnMemoryArenaChunk(memory_arena *arena, u64 chunk_size) {
    u64 current_position = GetMemoryArenaPosition(arena);
    u64 new_position = 0;

    if (chunk_size < current_position) {
        new_position = current_position - chunk_size;
    }

    RewindMemoryArenaPosition(arena, new_position);
}

#endif // DANI_LIB_MEMORY_ARENA_IMPLEMENTATION
