// System includes
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>
#include <intrin.h>

// Danilib includes
#include "danilib/dani_base.h"
#include "danilib/dani_arena.h"
#include "danilib/dani_string.h"

#define DEFAULT_MEMORY_COMMIT_SIZE KiB(8)
#define DEFAULT_MEMORY_RESERVE_SIZE GiB(32)
#define DEFAULT_MEMORY_ALIGNMENT 32

// Project includes
#include "base/console.c"
#include "base/arguments.c"
#include "base/file.c"
#include "base/profiler.c"
#include "base/random.c"
#include "base/hashmap.c"

// Danilib implementations
#define DANI_LIB_MEMORY_ARENA_IMPLEMENTATION
#define DANI_LIB_MEMORY_ARENA_PANIC
static void MemoryArenaPanic(void) {
    FatalWindowsError(GetLastError());
}

#include "danilib/dani_arena.h"
#undef DANI_LIB_MEMORY_ARENA_IMPLEMENTATION

#define DANI_LIB_STRING_IMPLEMENTATION
#include "danilib/dani_string.h"
#undef DANI_LIB_STRING_IMPLEMENTATION

#define EXIT_CODE_SUCCESS 0
#define EXIT_CODE_MISSING_ARGUMENT (0x20000000 + 2)
#define EXIT_CODE_INVALID_ARGUMENT (0x20000000 + 3)
#define EXIT_CODE_INVALID_ATTEMPT (0x20000000 + 4)

#include "attempts/common.c"
#include "attempts/attempts.c"

#include "generator.c"
#include "benchmark.c"
#include "program.c"

b32 StringToU32(string value, u32 *out_result);

void ProgramStart(void) {
    SetUpConsole();
    InitRandom();
    
    memory_arena *arena = MemoryArena(
        DEFAULT_MEMORY_RESERVE_SIZE,
        DEFAULT_MEMORY_COMMIT_SIZE,
        DEFAULT_MEMORY_ALIGNMENT);

    u32 exit_code;

    Unused(GetNextCommandLineArgument()); // Skip executable name
    string argument = GetNextCommandLineArgument();
    if (argument.length > 0) {
        if (IsStringEqual(argument, StringLiteral("generate"))) {
            exit_code = RunGenerator(arena);
        } else if (IsStringEqual(argument, StringLiteral("benchmark"))) {
            exit_code = RunBenchmark(arena);
        } else if (IsStringEqual(argument, StringLiteral("run"))) {
            argument = GetNextCommandLineArgument();

            u32 index;
            b32 is_index = StringToU32(argument, &index);
            if (IsFalse(is_index)) {
                index = ATTEMPT_FASTEST;
            }
            exit_code = RunProgram(arena, index);
        } else {
            exit_code = EXIT_CODE_INVALID_ARGUMENT;
            string error = StringFormat(arena, "Invalid argument \"%.*s\" - Has to be \"generate\", \"benchmark\", or \"run\".\n", StringToFormatArgs(argument));
            WriteError(error);
        }
    } else {
        exit_code = EXIT_CODE_MISSING_ARGUMENT;
        string error = StringFormat(arena, "Missing argument - Has to be \"generate\", \"benchmark\", or \"run\".\n", StringToFormatArgs(argument));
        WriteError(error);
    }

    ExitProcess(exit_code); // TODO: Maybe use terminate process instead?
}

b32 StringToU32(string value, u32 *out_result) {
    b32 result = value.length > 0;
    u32 parsed_value = 0;

    for (u32 i = 0; i < value.length && result; i++) {
        u8 digit = value.base[i] - '0';
        b32 is_digit = digit < 10;

        if (is_digit) {
            parsed_value = (parsed_value * 10) + digit;
        } else {
            parsed_value = 0;
            result = B32_FAILURE;
        }
    }

    *out_result = parsed_value;
    return (result);
}